#######################################################################
# Copyright (c) 2017 Josef Spidlen, Ph.D., FlowJo, LLC
#
# License
# The software is distributed under the terms of the 
# Artistic License 2.0
# http://www.r-project.org/Licenses/Artistic-2.0
# 
# Disclaimer
# This software and documentation come with no warranties of any kind.
# This software is provided "as is" and any express or implied 
# warranties, including, but not limited to, the implied warranties of
# merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the  copyright holder be liable for any direct, 
# indirect, incidental, special, exemplary, or consequential damages
# (including but not limited to, procurement of substitute goods or 
# services; loss of use, data or profits; or business interruption)
# however caused and on any theory of liability, whether in contract,
# strict liability, or tort arising in any way out of the use of this 
# software.
######################################################################

## This will force the creation of a backup input and output CSV file for additional debugging
debug_input_output <- FALSE

suppressMessages(library("FlowSOM"))
suppressMessages(library("flowCore"))

## Mainly we want FlowSOM >= 1.8.0 (seed only works for FlowSOM >= 1.8.0)
## Tested with FlowSOM >= 1.8.0 and FlowSOM >= 1.10.0
## Also, having pheatmap (1.0.8) is desirable in order to produce some additional (optional) graphics
sessionInfo()

## Just for debugging to see what FlowJo is actually passing to the script.
##
## cat("FJ-DATA-FILE-PATH            = FJ_DATA_FILE_PATH \n")
## cat("FJ-CSV-OUPUT-FILE            = FJ_CSV_OUPUT_FILE \n")
## cat("FJ-PAR-NUM-CLUST             = FJ_PAR_NUM_CLUST \n")
## cat("FJ-PARAMS-LIST               = FJ_PARAMS_LIST \n")
## cat("FJ-BG-META-CLUSTERS          = FJ_BG_META_CLUSTERS \n")
## cat("FJ-PLOT-OPTION               = FJ_PLOT_OPTION \n")
## cat("FJ-VIEW-OPTION               = FJ_VIEW_OPTION \n")
## cat("FJ-PLOT-ALL                  = FJ_PLOT_ALL \n")
## cat("FJ-PAR-SCALE                 = FJ_PAR_SCALE \n")
## cat("FJ-PAR-XDIM                  = FJ_PAR_XDIM \n")
## cat("FJ-PAR-YDIM                  = FJ_PAR_YDIM \n")
## cat("FJ-PAR-CREATE-PDF            = FJ_PAR_CREATE_PDF \n")
## cat("FJ-PAR-NODE-SCALE            = FJ_PAR_NODE_SCALE \n")
## cat("FJ-PAR-FIXED-NODE-SIZE       = FJ_PAR_FIXED_NODE_SIZE \n")
## cat("FJ-PAR-APPLY-ON-PREV         = FJ_PAR_APPLY_ON_PREV \n")
## cat("FJ-PAR-ADD-CELLIDS-TO-RESULT = FJ_PAR_ADD_CELLIDS_TO_RESULT \n")

plotAllAsStars <- FJ_PLOT_ALL
bgMetaClusters <- FJ_BG_META_CLUSTERS
plotView <- "FJ_VIEW_OPTION"
plotPar <- "FJ_PLOT_OPTION"
EventNumberDP <- "EventNumberDP"
CellIdDP <- "CellId"
addCellIdsToResult <- FJ_PAR_ADD_CELLIDS_TO_RESULT
runningInFlowJo <- !addCellIdsToResult

## Generates GUID like "47689979-e471-47a6-aa1a-7910d2d42b03"
## As a side effect, this also sets the seed based in the data passed in
generateGUID <- function(data) {
	mySeed <- round(1000*mean(as.matrix(data))) + dim(data)[1] + dim(data)[2]
	while (mySeed > .Machine$integer.max) mySeed <- round(mySeed/3)
	set.seed(as.integer(mySeed))
	
	part1 <- paste(unlist(lapply(1:8, function(x) { sprintf("%x", round(runif(1, 0, 9))) })), collapse="", sep="")
	part2 <- paste(unlist(lapply(1:4, function(x) { sprintf("%x", round(runif(1, 0, 15))) })), collapse="", sep="")
	part3 <- paste(unlist(lapply(1:4, function(x) { sprintf("%x", round(runif(1, 0, 15))) })), collapse="", sep="")
	part4 <- paste(unlist(lapply(1:4, function(x) { sprintf("%x", round(runif(1, 0, 15))) })), collapse="", sep="")
	part5 <- paste(unlist(lapply(1:12, function(x) { sprintf("%x", round(runif(1, 0, 15))) })), collapse="", sep="")
	paste(part1, part2, part3, part4, part5, sep="-") 
}

## Returns a list of 2 elements
## fcs = flowFrame created from a CSV file as exported by FlowJo
## cellIds = dataFrame with CellIds (if available)
##
## Requires library("flowCore")
##
## csvPath: path the the CSV file as exported from FlowJo (including header with both stain and parameter is recommended when exporting from FlowJo)
## dropCompPrefix: FlowJo indicates that a patameter is compensated by the Comp- prefix in parameter names, we can drop that by setting dropCompPrefix to TRUE if we don't want it
flowFrameFromFlowJoCSV <- function(csvPath, dropCompPrefix=FALSE) {
	data <- read.csv(csvPath, check.names=FALSE)
	cellIdsColumn <- rep(NA, nrow(data))
	cellIds <- data.frame(CellIdDP = cellIdsColumn)
	names(cellIds) <- CellIdDP
	if (CellIdDP %in% names(data)) {
		cellIds <- data[, CellIdDP]
		data <- data[, -which(names(data) %in% c(CellIdDP))]
	}

	colNames <- colnames(data)
	npar <- length(colNames)
	id <- paste("$P",1:npar,sep="")

	maxRange <- unlist(lapply(1:npar, function(x) {
		max(data[,x])
	}))
	range <- maxRange + 1
	minRange <- unlist(lapply(1:npar, function(x) {
		min(0, min(data[,x]))
	}))
	
	colNames <- gsub("^\\s+|\\s+$", "", colNames) 	# Trim the leading and tailing whitespaces

	## FlowJo splits parameter names and description by " :: ", e.g., "Comp-PE-Cy5-A :: CD19#PE-Cy5"
	names <- unlist(lapply(1:npar, function(i) {
		unlist(strsplit(colNames[i], " :: "))[1]
	}))
	## FlowJo indicates that a patameter is compensated by the Comp- prefix in its name, we can drop that if undesirable
	if (dropCompPrefix) {
		names <- gsub("^\\Comp-", "", names)
	}
	## Again splitting by " :: " but this time taking the second part (if present) as description
	desc <- unlist(lapply(1:npar, function(i) {
		s <- unlist(strsplit(colNames[i], " :: "))
		if (length(s) >= 2) s[2]
		else NA
	}))
	
	pars <- new("AnnotatedDataFrame",
		data=data.frame(
				row.names=I(id), name=I(names), desc=I(desc),
				range=range, minRange=minRange, maxRange=maxRange),
		varMetadata=data.frame(
				row.names=I(c("name","desc","range", "minRange", "maxRange")),
				labelDescription=I(c("Name of Parameter", "Description of Parameter", "Range of Parameter", "Minimum Parameter Value after Transforamtion", "Maximum Parameter Value after Transformation"))
		)
	)

	## Build the descripton
	txt <- list()
	txt["FCSversion"] <- "3"
	txt["$PAR"] <- as.character(npar)
	txt["$TOT"] <- as.character(dim(data)[1])
	txt["$BYTEORD"] <- "4,3,2,1"
	txt["$DATATYPE"] <- "F"
	txt["$MODE"] <- "L"
	txt["$NEXTDATA"] <- "0"
	
	txt["$BEGINANALYSIS"] <- "0"
	txt["$ENDANALYSIS"] <- "0"
	txt["$BEGINSTEXT"] <- "0"
	txt["$ENDSTEXT"] <- "0"
	dataStart <- 1000 ## This doesn't really matter, it's not an FCS file anyway, but let's provide something reasonable
	txt["$BEGINDATA"] <- as.character(dataStart)
	txt["$ENDDATA"] <- as.character(dataStart + npar * dim(data)[1] * 4 - 1) ## $DATATYPE is F and PnB is 32, so here we do: start + rows x cols x 4 bytes per value - 1

	txt[paste(id,"N",sep="")] <- names
	txt[paste(id,"S",sep="")] <- desc
	txt[paste(id,"R",sep="")] <- as.character(range)
	txt[paste(id,"B",sep="")] <- "32"
	txt[paste(id,"E",sep="")] <- "0,0"
	txt[paste(id,"G",sep="")] <- "1"
	txt[paste("flowCore_",id,"Rmin",sep="")] <- as.character(minRange)
	txt[paste("flowCore_",id,"Rmax",sep="")] <- as.character(maxRange)
	
	txt["$FIL"] <- basename(csvPath)
	txt["FILENAME"] <- csvPath
	txt["$SYS"] <- paste(Sys.info()['sysname'], Sys.info()['release'])
	txt["CREATOR"] <- R.version.string
	txt["transformation"] <- "applied"
	txt["GUID"] <- generateGUID(data)
	txt["ORIGINALGUID"] <- generateGUID(data)
	
	txt["$INST"] <- "Unknown institution"
	txt["$OP"] <- "Unknown operator"
	txt["$CYT"] <- "Unknown cytometer"
	txt["$SRC"] <- "Unknown source"
	
	curtime <- Sys.time()
	txt["$DATE"] <- toupper(format(curtime, "%d-%b-%Y")) ## Current data in the format of "05-OCT-2017"
	timeDelta <- 120 ## picking 2 minutes as cytometer run time
	if ("time" %in% tolower(names)) {
		txt["$TIMESTEP"] <- "0.01" ## We don't know that, so just something that would be reasonable-ish
		timeIndex <- which(tolower(names) == "time")
		timeDelta <- ceiling((max(data[,timeIndex]) - min(data[,timeIndex])) * 0.01)
	} 
	if (toupper(format(curtime - timeDelta, "%d-%b-%Y")) == toupper(format(curtime, "%d-%b-%Y"))) {
		## Subtracting timeDelta is still the same day, so let's do it that way.
		txt["$BTIM"] <- format(curtime - timeDelta, "%H:%M:%S")
		txt["$ETIM"] <- format(curtime, "%H:%M:%S")
	} else {
		## Subtracting timeDelta gets us to yesterday, that's not good, so let's just say we started at midnight and add delta to it as ETIM... 
		## (all this is arbitrary anyway, we just want resonable values)
		txt["$BTIM"] <- "00:00:00"
		txt["$ETIM"] <- sprintf("%02i:%02i:%02i", floor(timeDelta / 3600), floor((timeDelta%%3600) / 60), floor(timeDelta%%60))
	}
	colnames(data) <- names
	list(fcs=new("flowFrame", exprs=as.matrix(data), parameters=pars, description=txt), cellIds=cellIds)	
}



plotFlowSOMResults <- function(x, shift=1)
{
	
	if (plotAllAsStars) {
		## Use PlotStars
		if (bgMetaClusters) {
			PlotStars(
					x$FlowSOM,
					view=plotView,
					backgroundValues = as.factor(as.numeric(x$metaclustering) - shift)) ## Used in FlowJo as -1 to make it go 0..n-1 instead 1..n to match FlowJo's populations
		} else {
			PlotStars(
					x$FlowSOM,
					view=plotView)
		}
	} else {
		## Use PlotMarker
		if (bgMetaClusters) {
			PlotMarker(
					x$FlowSOM,
					plotPar,
					view=plotView,
					backgroundValues = as.factor(as.numeric(x$metaclustering) - shift)) ## Used in FlowJo as -1 to make it go 0..n-1 instead 1..n to match FlowJo's populations
		} else {
			PlotMarker(
					x$FlowSOM,
					plotPar,
					view=plotView)
		}	
	}
}

if (debug_input_output)
	file.copy("FJ_DATA_FILE_PATH", paste("FJ_DATA_FILE_PATH", "backup.csv", sep="."))

tmpRet = flowFrameFromFlowJoCSV("FJ_DATA_FILE_PATH");
inputFCS <- tmpRet$fcs
cellIds <- as.matrix(tmpRet$cellIds)
rm(tmpRet)

eventsCount <- length(inputFCS@exprs[,1])

parNames <- c(FJ_PARAMS_LIST)

## In CSV files, the parameter names are often like FJComp-xxx while in parNames we may be getting Comp-xxx
if (any(grepl("FJComp", parNames)) && !any(grepl("FJComp", colnames(inputFCS)))) {
	## If it looks like there is FJComp-xxx in parNames but no FJComp-xxx in the column names of the FCS file, then
	## rename FJComp-xxx to Comp-xxx in the parNames and we will be looking for those instead.
	parNames <- gsub("^\\FJComp-", "Comp-", parNames)
}

if (length(parNames) == 0)
	stop("Some input parameters need to be selected!", call.=FALSE)

# FlowJo sometimes gives us names like [V705-A] but the actual names in the FCS file are <V705-A>
# FlowJo also replaces '_' with '/' so let's try fix that as well
parNames <- unlist(lapply(parNames, function(name) {
					if (name %in% colnames(inputFCS)) {
						name
					} else {
						# Let's try [] to <>
						name2 <- gsub("[", "<", name, fixed=TRUE)
						name2 <- gsub("]", ">", name2, fixed=TRUE)
						if (name2 %in% colnames(inputFCS)) {
							name2 # Worked, return it
						} else {
							# Previous fix did not do it, _ => / on the original names
							name2 <- gsub("_", "/", name, fixed=TRUE)
							if (name2 %in% colnames(inputFCS)) {
								name2 # Worked, return it
							} else {
								# That did not work either, let's try both [] => on top of the previous fix (_ => /)
								name3 <- gsub("[", "<", name2, fixed=TRUE)
								name3 <- gsub("]", ">", name3, fixed=TRUE)
								if (name3 %in% colnames(inputFCS)) {
									name3 # Worked, finally, return it
								} else {
									## Maybe we read a wrong dataset?
									cat(paste("The input FCS file does not contain the provided input parameter, missing", name, "\n"))
									"MISSINGPARAMETER"
								}
							}
						}
					}
				}))

## DEBUG info
inputFCS
parNames
eventsCount
summary(cellIds)

if (length(parNames) == 0 || "MISSINGPARAMETER" %in% parNames)
	stop("The input file is missing some of the specified parameters.", call.=FALSE)
if (eventsCount == 0)
	stop("R failed to read the input file.", call.=FALSE)


if (!plotAllAsStars) {
	## If not plotting all then make sure plotPar is among parameters in the FCS file
	if (!(plotPar %in% colnames(inputFCS))) {
		plotPar2 <- gsub("[", "<", plotPar, fixed=TRUE)
		plotPar2 <- gsub("]", ">", plotPar2, fixed=TRUE)
		if (plotPar2 %in% colnames(inputFCS)) {
			plotPar <- plotPar2
		} else {
			plotPar2 <- gsub("_", "/", plotPar, fixed=TRUE)
			if (plotPar2 %in% colnames(inputFCS)) {
				plotPar <- plotPar2
			} else {
				plotPar3 <- gsub("[", "<", plotPar2, fixed=TRUE)
				plotPar3 <- gsub("]", ">", plotPar3, fixed=TRUE)
				if (plotPar3 %in% colnames(inputFCS)) {
					plotPar <- plotPar3
				} else {
					stop(paste("The input FCS file does not contain ", plotPar), call.=FALSE)
				}
			} 
		}
	}
}

parIndices <- match(parNames, colnames(inputFCS))
if (length(parNames) == 0 || length(parIndices) == 0)
	stop("Something seems wrong, it's like the input FCS file does not contain the provided input parameters.", call.=FALSE)

if (nchar("FJ_PAR_APPLY_ON_PREV") > 5) { ## Expected either "None" or a valid file path
  load("FJ_PAR_APPLY_ON_PREV")
  fSOM$FlowSOM <- NewData(fSOM$FlowSOM, inputFCS);
} else {
  fSOM <- FlowSOM(
    inputFCS,
    compensate = FALSE,
    transform = FALSE,
    scale = FJ_PAR_SCALE,
    colsToUse = parIndices, 
    xdim = FJ_PAR_XDIM, 
    ydim = FJ_PAR_YDIM,
    nClus = FJ_PAR_NUM_CLUST,
    seed = 12345)
}


## Code to extract MFIs

## Provide an empty matrix to store MFIs
n_Cl <- fSOM$FlowSOM$map$nNodes
n_Meta <- length(levels(fSOM$metaclustering))
usefulCols1 <- colnames(inputFCS)[colnames(inputFCS) != EventNumberDP]
usefulCols2 <- colnames(inputFCS)[colnames(inputFCS) != CellIdDP]
usefulCols <- intersect(usefulCols1, usefulCols2)
usefulParIndices <- match(usefulCols, colnames(inputFCS))

cluster_MFI <- matrix(NA, nrow = n_Cl, ncol = ncol(inputFCS), dimnames = list(paste0("Cl", seq_len(n_Cl)), colnames(inputFCS)))
meta_MFI <- matrix(NA, nrow = n_Meta, ncol = ncol(inputFCS), dimnames = list(paste0("Pop", seq_len(n_Meta)), colnames(inputFCS)))

## Compute the MFIs
for (marker in colnames(inputFCS)) {
	mfis <- tapply(exprs(inputFCS)[,marker], fSOM$FlowSOM$map$mapping[,1], median)
	cluster_MFI[paste0("Cl", names(mfis)), marker] <- mfis
	
	mfis <- tapply(exprs(inputFCS)[,marker], fSOM$metaclustering[fSOM$FlowSOM$map$mapping[,1]], median)
	meta_MFI[paste0("Pop", names(mfis)), marker] <- mfis
}

colnames(cluster_MFI) <- fSOM$FlowSOM$prettyColnames[colnames(cluster_MFI)]
colnames(meta_MFI) <- fSOM$FlowSOM$prettyColnames[colnames(meta_MFI)]
rownames(meta_MFI) <- paste0("Pop", seq_len(n_Meta) - 1) ## Start from 0 to match FlowJo's population names

write.csv(cluster_MFI[,usefulParIndices], paste0("FJ_CSV_OUPUT_FILE", ".ClMFI.csv"), row.names=FALSE)
write.csv(meta_MFI[,usefulParIndices], paste0("FJ_CSV_OUPUT_FILE", ".PopMFI.csv"), row.names=FALSE)

## Extract metaclustering to pass back to FlowJo (or SeqGeq)
labels <- NULL
shiftIds <- 1;
if (runningInFlowJo) {
	## FlowJo, let's do labels as 100, 200, 300, all with a tiny bit of noise (to make FlowJo cluster those better)
	labels <- as.matrix(as.numeric((fSOM$metaclustering[fSOM$FlowSOM$map$mapping[,1]])) * 100 + runif(min=-0.01, max=0.01, n=eventsCount));
} else {
	## SeqGeq, let's just keep the labels as 1, 2, 3, ... 
	labels <- as.matrix(as.numeric((fSOM$metaclustering[fSOM$FlowSOM$map$mapping[,1]])));
	shiftIds <- 0;
}
colnames(labels) <- c("FlowSOM")

if (addCellIdsToResult && !(NA %in% cellIds)) {
	labels <- cbind(cellIds, labels)
	colnames(labels) <- c(CellIdDP, "FlowSOM")
}

write.csv(labels, file="FJ_CSV_OUPUT_FILE", row.names=FALSE, quote=FALSE)
if (debug_input_output)
	file.copy("FJ_CSV_OUPUT_FILE", paste("FJ_CSV_OUPUT_FILE", "backup.csv", sep="."))

## 15 is the default maxNodeSize, we provide a percentage based scaling factor
fSOM$FlowSOM <- UpdateNodeSize(fSOM$FlowSOM, maxNodeSize=(15*0.01*FJ_PAR_NODE_SCALE), reset=FJ_PAR_FIXED_NODE_SIZE)

png(file=paste0("FJ_CSV_OUPUT_FILE", ".png"), bg="transparent", width = 1280, height = 1280, units = "px", pointsize = 24)
plotFlowSOMResults(fSOM, shiftIds)
dev.off()

if(FJ_PAR_CREATE_PDF)
{
  pdf(file=paste0("FJ_CSV_OUPUT_FILE", ".pdf"), bg="transparent", useDingbats=FALSE)
  plotFlowSOMResults(fSOM, shiftIds)
  dev.off()
  if (require("pheatmap") == TRUE) {
	  ## Additional graphics (heatmaps) if there is a pheatmap package installed.
	  ## Also, wrapping into tryCatch is hclust has been seen to fail here and we don't want it to take the whole script down.
	  tryCatch({
		  pheatmap::pheatmap(cluster_MFI[,usefulParIndices], scale = "none", fontsize=5, file=paste0("FJ_CSV_OUPUT_FILE", ".ClHm.pdf"), bg="transparent", useDingbats=FALSE)
		  pheatmap::pheatmap(meta_MFI[,usefulParIndices], scale = "none", file=paste0("FJ_CSV_OUPUT_FILE", ".PopHm.pdf"), bg="transparent", useDingbats=FALSE)
	    }, error = function(e) {
			writeLines("Warning, pheatmap call failed, some heatmap figures will not be created.")
		})
  } else {
	  writeLines("Warning, the pheatmap package is not present, some heatmap figures will not be created.")
  }
}

save(fSOM, file=paste0("FJ_CSV_OUPUT_FILE", ".RData"))
write.csv(parNames, paste0("FJ_CSV_OUPUT_FILE", ".pars.csv"), row.names=FALSE)

# R seems to be saving the .RData when exiting, so let's clean up to at least make that tiny (i.e., empty environment)
rm(list=ls())
