#######################################################################
# Copyright (c) 2017 Josef Spidlen, Ph.D., FlowJo, LLC
#
# License
# The software is distributed under the terms of the 
# Artistic License 2.0
# http://www.r-project.org/Licenses/Artistic-2.0
# 
# Disclaimer
# This software and documentation come with no warranties of any kind.
# This software is provided "as is" and any express or implied 
# warranties, including, but not limited to, the implied warranties of
# merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the  copyright holder be liable for any direct, 
# indirect, incidental, special, exemplary, or consequential damages
# (including but not limited to, procurement of substitute goods or 
# services; loss of use, data or profits; or business interruption)
# however caused and on any theory of liability, whether in contract,
# strict liability, or tort arising in any way out of the use of this 
# software.
######################################################################

suppressMessages(library("FlowSOM"))
suppressMessages(library("flowCore"))

sessionInfo()

plotAllAsStars <- TRUE
bgMetaClusters <- TRUE
plotView <- "MST"
plotPar <- "All as piecharts"
EventNumberDP <- "EventNumberDP"
CellIdDP <- "CellId"
addCellIdsToResult <- TRUE
runningInFlowJo <- !addCellIdsToResult

f1_score <- function(predicted, expected) {
	predicted <- as.factor(predicted)
	expected  <- as.factor(expected)
	cm = as.matrix(table(expected, predicted))	
	precision <- diag(cm) / colSums(cm)
	recall <- diag(cm) / rowSums(cm)
	f1 <- ifelse(precision + recall == 0, 0, 2 * precision * recall / (precision + recall))	
	f1[is.na(f1)] <- 0
	f1
}

## Generates GUID like "47689979-e471-47a6-aa1a-7910d2d42b03"
## As a side effect, this also sets the seed based in the data passed in
generateGUID <- function(data) {
	mySeed <- round(1000*mean(as.matrix(data))) + dim(data)[1] + dim(data)[2]
	while (mySeed > .Machine$integer.max) mySeed <- round(mySeed/3)
	set.seed(as.integer(mySeed))
	
	part1 <- paste(unlist(lapply(1:8, function(x) { sprintf("%x", round(runif(1, 0, 9))) })), collapse="", sep="")
	part2 <- paste(unlist(lapply(1:4, function(x) { sprintf("%x", round(runif(1, 0, 15))) })), collapse="", sep="")
	part3 <- paste(unlist(lapply(1:4, function(x) { sprintf("%x", round(runif(1, 0, 15))) })), collapse="", sep="")
	part4 <- paste(unlist(lapply(1:4, function(x) { sprintf("%x", round(runif(1, 0, 15))) })), collapse="", sep="")
	part5 <- paste(unlist(lapply(1:12, function(x) { sprintf("%x", round(runif(1, 0, 15))) })), collapse="", sep="")
	paste(part1, part2, part3, part4, part5, sep="-") 
}

## Returns a list of 2 elements
## fcs = flowFrame created from a CSV file as exported by FlowJo
## cellIds = dataFrame with CellIds (if available)
##
## Requires library("flowCore")
##
## csvPath: path the the CSV file as exported from FlowJo (including header with both stain and parameter is recommended when exporting from FlowJo)
## dropCompPrefix: FlowJo indicates that a patameter is compensated by the Comp- prefix in parameter names, we can drop that by setting dropCompPrefix to TRUE if we don't want it
flowFrameFromFlowJoCSV <- function(csvPath, dropCompPrefix=FALSE) {
	data <- read.csv(csvPath, check.names=FALSE)
	cellIdsColumn <- rep(NA, nrow(data))
	cellIds <- data.frame(CellIdDP = cellIdsColumn)
	names(cellIds) <- CellIdDP
	if (CellIdDP %in% names(data)) {
		cellIds <- data[, CellIdDP]
		data <- data[, -which(names(data) %in% c(CellIdDP))]
	}
	
	colNames <- colnames(data)
	npar <- length(colNames)
	id <- paste("$P",1:npar,sep="")
	
	maxRange <- unlist(lapply(1:npar, function(x) {
						max(data[,x])
					}))
	range <- maxRange + 1
	minRange <- unlist(lapply(1:npar, function(x) {
						min(0, min(data[,x]))
					}))
	
	colNames <- gsub("^\\s+|\\s+$", "", colNames) 	# Trim the leading and tailing whitespaces
	
	## FlowJo splits parameter names and description by " :: ", e.g., "Comp-PE-Cy5-A :: CD19#PE-Cy5"
	names <- unlist(lapply(1:npar, function(i) {
						unlist(strsplit(colNames[i], " :: "))[1]
					}))
	## FlowJo indicates that a patameter is compensated by the Comp- prefix in its name, we can drop that if undesirable
	if (dropCompPrefix) {
		names <- gsub("^\\Comp-", "", names)
	}
	## Again splitting by " :: " but this time taking the second part (if present) as description
	desc <- unlist(lapply(1:npar, function(i) {
						s <- unlist(strsplit(colNames[i], " :: "))
						if (length(s) >= 2) s[2]
						else NA
					}))
	
	pars <- new("AnnotatedDataFrame",
			data=data.frame(
					row.names=I(id), name=I(names), desc=I(desc),
					range=range, minRange=minRange, maxRange=maxRange),
			varMetadata=data.frame(
					row.names=I(c("name","desc","range", "minRange", "maxRange")),
					labelDescription=I(c("Name of Parameter", "Description of Parameter", "Range of Parameter", "Minimum Parameter Value after Transforamtion", "Maximum Parameter Value after Transformation"))
			)
	)
	
	## Build the descripton
	txt <- list()
	txt["FCSversion"] <- "3"
	txt["$PAR"] <- as.character(npar)
	txt["$TOT"] <- as.character(dim(data)[1])
	txt["$BYTEORD"] <- "4,3,2,1"
	txt["$DATATYPE"] <- "F"
	txt["$MODE"] <- "L"
	txt["$NEXTDATA"] <- "0"
	
	txt["$BEGINANALYSIS"] <- "0"
	txt["$ENDANALYSIS"] <- "0"
	txt["$BEGINSTEXT"] <- "0"
	txt["$ENDSTEXT"] <- "0"
	dataStart <- 1000 ## This doesn't really matter, it's not an FCS file anyway, but let's provide something reasonable
	txt["$BEGINDATA"] <- as.character(dataStart)
	txt["$ENDDATA"] <- as.character(dataStart + npar * dim(data)[1] * 4 - 1) ## $DATATYPE is F and PnB is 32, so here we do: start + rows x cols x 4 bytes per value - 1
	
	txt[paste(id,"N",sep="")] <- names
	txt[paste(id,"S",sep="")] <- desc
	txt[paste(id,"R",sep="")] <- as.character(range)
	txt[paste(id,"B",sep="")] <- "32"
	txt[paste(id,"E",sep="")] <- "0,0"
	txt[paste(id,"G",sep="")] <- "1"
	txt[paste("flowCore_",id,"Rmin",sep="")] <- as.character(minRange)
	txt[paste("flowCore_",id,"Rmax",sep="")] <- as.character(maxRange)
	
	txt["$FIL"] <- basename(csvPath)
	txt["FILENAME"] <- csvPath
	txt["$SYS"] <- paste(Sys.info()['sysname'], Sys.info()['release'])
	txt["CREATOR"] <- R.version.string
	txt["transformation"] <- "applied"
	txt["GUID"] <- generateGUID(data)
	txt["ORIGINALGUID"] <- generateGUID(data)
	
	txt["$INST"] <- "Unknown institution"
	txt["$OP"] <- "Unknown operator"
	txt["$CYT"] <- "Unknown cytometer"
	txt["$SRC"] <- "Unknown source"
	
	curtime <- Sys.time()
	txt["$DATE"] <- toupper(format(curtime, "%d-%b-%Y")) ## Current data in the format of "05-OCT-2017"
	timeDelta <- 120 ## picking 2 minutes as cytometer run time
	if ("time" %in% tolower(names)) {
		txt["$TIMESTEP"] <- "0.01" ## We don't know that, so just something that would be reasonable-ish
		timeIndex <- which(tolower(names) == "time")
		timeDelta <- ceiling((max(data[,timeIndex]) - min(data[,timeIndex])) * 0.01)
	} 
	if (toupper(format(curtime - timeDelta, "%d-%b-%Y")) == toupper(format(curtime, "%d-%b-%Y"))) {
		## Subtracting timeDelta is still the same day, so let's do it that way.
		txt["$BTIM"] <- format(curtime - timeDelta, "%H:%M:%S")
		txt["$ETIM"] <- format(curtime, "%H:%M:%S")
	} else {
		## Subtracting timeDelta gets us to yesterday, that's not good, so let's just say we started at midnight and add delta to it as ETIM... 
		## (all this is arbitrary anyway, we just want resonable values)
		txt["$BTIM"] <- "00:00:00"
		txt["$ETIM"] <- sprintf("%02i:%02i:%02i", floor(timeDelta / 3600), floor((timeDelta%%3600) / 60), floor(timeDelta%%60))
	}
	colnames(data) <- names
	list(fcs=new("flowFrame", exprs=as.matrix(data), parameters=pars, description=txt), cellIds=cellIds)	
}



plotFlowSOMResults <- function(x, shift=1)
{
	
	if (plotAllAsStars) {
		## Use PlotStars
		if (bgMetaClusters) {
			PlotStars(
					x$FlowSOM,
					view=plotView,
					backgroundValues = as.factor(as.numeric(x$metaclustering) - shift)) ## Used in FlowJo as -1 to make it go 0..n-1 instead 1..n to match FlowJo's populations
		} else {
			PlotStars(
					x$FlowSOM,
					view=plotView)
		}
	} else {
		## Use PlotMarker
		if (bgMetaClusters) {
			PlotMarker(
					x$FlowSOM,
					plotPar,
					view=plotView,
					backgroundValues = as.factor(as.numeric(x$metaclustering) - shift)) ## Used in FlowJo as -1 to make it go 0..n-1 instead 1..n to match FlowJo's populations
		} else {
			PlotMarker(
					x$FlowSOM,
					plotPar,
					view=plotView)
		}	
	}
}

##########################################################################################################################################################################
##########################################################################################################################################################################
## TODO: This needs to be adjusted as necessary
tmpRet <- flowFrameFromFlowJoCSV("/Users/josef/Library/Caches/com.flowjo.flowjovx/SeqGeqRemoteFileCache/zip/22187268517b81a7ee4afd05254033bb-ILC.geqzip/FlowSOM/ILC_FACS_CombatZeroFix.csv..ExtNode.csv.backup.csv");
parNames <- c("ILC_Phenotype","Patient","PCAFL__1","PCAFL__2","PCAFL__3","PCAFL__4","PCAFL__5","PCAFL__6","PCAFL__7")
xdim <- 8 
ydim <- 8
nClus <- 4
refCol <- c("ILC_Phenotype")
##########################################################################################################################################################################
##########################################################################################################################################################################

inputFCS <- tmpRet$fcs
cellIds <- as.matrix(tmpRet$cellIds)
rm(tmpRet)

eventsCount <- length(inputFCS@exprs[,1])


## In CSV files, the parameter names are often like FJComp-xxx while here we are getting Comp-xxx
parNames <- gsub("^\\FJComp-", "Comp-", parNames)

if (length(parNames) == 0)
	stop("Some input parameters need to be selected!", call.=FALSE)

parNames <- unlist(lapply(parNames, function(name) {
					if (name %in% colnames(inputFCS)) {
						name
					} else {
						# Let's try [] to <>
						name2 <- gsub("[", "<", name, fixed=TRUE)
						name2 <- gsub("]", ">", name2, fixed=TRUE)
						if (name2 %in% colnames(inputFCS)) {
							name2 # Worked, return it
						} else {
							# Previous fix did not do it, _ => / on the original names
							name2 <- gsub("_", "/", name, fixed=TRUE)
							if (name2 %in% colnames(inputFCS)) {
								name2 # Worked, return it
							} else {
								# That did not work either, let's try both [] => on top of the previous fix (_ => /)
								name3 <- gsub("[", "<", name2, fixed=TRUE)
								name3 <- gsub("]", ">", name3, fixed=TRUE)
								if (name3 %in% colnames(inputFCS)) {
									name3 # Worked, finally, return it
								} else {
									## Maybe we read a wrong dataset?
									cat(paste("The input FCS file does not contain the provided input parameter, missing", name, "\n"))
									"MISSINGPARAMETER"
								}
							}
						}
					}
				}))

## DEBUG info
inputFCS
parNames
eventsCount
summary(cellIds)

if (length(parNames) == 0 || "MISSINGPARAMETER" %in% parNames) 
	stop("The input file is missing some of the specified parameters.", call.=FALSE)
if (eventsCount == 0)
	stop("R failed to read the input file.", call.=FALSE)


if (!plotAllAsStars) {
	## If not plotting all then make sure plotPar is among parameters in the FCS file
	if (!(plotPar %in% colnames(inputFCS))) {
		plotPar2 <- gsub("[", "<", plotPar, fixed=TRUE)
		plotPar2 <- gsub("]", ">", plotPar2, fixed=TRUE)
		if (plotPar2 %in% colnames(inputFCS)) {
			plotPar <- plotPar2
		} else {
			plotPar2 <- gsub("_", "/", plotPar, fixed=TRUE)
			if (plotPar2 %in% colnames(inputFCS)) {
				plotPar <- plotPar2
			} else {
				plotPar3 <- gsub("[", "<", plotPar2, fixed=TRUE)
				plotPar3 <- gsub("]", ">", plotPar3, fixed=TRUE)
				if (plotPar3 %in% colnames(inputFCS)) {
					plotPar <- plotPar3
				} else {
					stop(paste("The input FCS file does not contain ", plotPar), call.=FALSE)
				}
			} 
		}
	}
}
## Temp, let's only use the PCA components.
parNames <- parNames[startsWith(parNames, "PCA")]

parIndices <- match(parNames, colnames(inputFCS))
if (length(parNames) == 0 || length(parIndices) == 0)
	stop("Something seems wrong, it's like the input FCS file does not contain the provided input parameters.", call.=FALSE)

##########################################################################################################################################################################
## FlowSOM loop for nClus
for(nClus in c(4:4)) {
fSOM <- FlowSOM(
	inputFCS, 
	compensate = FALSE, 
	transform = FALSE,
	scale = FALSE,
	colsToUse = parIndices, 
	xdim = xdim, 
	ydim = ydim,
	nClus = nClus,
	seed = 12345)

refInd <- match(refCol, colnames(inputFCS))
refLabels <- inputFCS@exprs[,refInd]
refLabelsTmp <- rep(0, length(refLabels))
for(i in 1:length(unique(refLabels))) {
	refLabelsTmp[refLabels == sort(unique(refLabels))[i]] <- i
}
refLabels <- as.matrix(refLabelsTmp, ncol=1)
rm(refLabelsTmp)
labels <- as.matrix(as.numeric((fSOM$metaclustering[fSOM$FlowSOM$map$mapping[,1]])));



##########################################################################################################################################################################
## FlowSOM results evaluation
resultList <- list()
print("FlowSOM results")
f1Vect <- c()
for(refI in 1:length(unique(refLabels))) {
	maxPopF1 <- -1
	MaxPopF1Index <- -1
	# Testing F1 for pop refI, which labels from clustering could that be?
	refLabelsTmp <- ifelse(refLabels == refI, 1, 2)
	for(i in 1:length(unique(labels))) {
		labelsTmp <- ifelse(labels == i, 1, 2)
		f1 <- f1_score(labelsTmp, refLabelsTmp)[1]
		if (f1 > maxPopF1) {
			maxPopF1 = f1
			MaxPopF1Index = i;
		}
		# print(paste("F1 for ref pop", refI, "based on cluster", i, "is", f1))
	}
	print(paste("With nClus =", nClus, "and xdim/ydim =", xdim, "best F1 for ref pop", refI, "is based on cluster", MaxPopF1Index, "and = ", maxPopF1))
	resultList[[refI]] = list(cluster=MaxPopF1Index, f1=maxPopF1)
	f1Vect <- cbind(f1Vect, maxPopF1)
}
print(paste("With nClus =", nClus, "FlowSOM results mean F1 =", mean(f1Vect)))
##########################################################################################################################################################################
} ## end of nClus look for FlowSOM



##########################################################################################################################################################################
## KMeans loop for nClus
for(nClus in 4:4) {
inputForKMeans <- inputFCS@exprs[,parIndices]
kmeansRes <- kmeans(inputForKMeans, nClus)

##########################################################################################################################################################################
## K-means results evaluation
resultListKM <- list()
print("K-means results")
f1VectKM <- c()
for(refI in 1:length(unique(refLabels))) {
	maxPopF1 <- -1
	MaxPopF1Index <- -1
	# Testing F1 for pop refI, which labels from clustering could that be?
	refLabelsTmp <- ifelse(refLabels == refI, 1, 2)
	for(i in 1:length(unique(kmeansRes$cluster))) {
		labelsTmp <- ifelse(kmeansRes$cluster == i, 1, 2)
		f1 <- f1_score(labelsTmp, refLabelsTmp)[1]
		if (f1 > maxPopF1) {
			maxPopF1 = f1
			MaxPopF1Index = i;
		}
		# print(paste("F1 for ref pop", refI, "based on cluster", i, "is", f1))
	}
	print(paste("With nClus =", nClus, "best F1 for ref pop", refI, "is based on cluster", MaxPopF1Index, "and = ", maxPopF1))
	resultListKM[[refI]] = list(cluster=MaxPopF1Index, f1=maxPopF1)
	f1VectKM <- cbind(f1VectKM, maxPopF1)
}
print(paste("With nClus =", nClus, "K-means results mean F1 =", mean(f1VectKM)))
##########################################################################################################################################################################
} ## end of nClus look for kMeans


save(list = ls(all.names = TRUE), file="ILCFlow.RData")
savehistory(file = "ILCFlow.RHistory")



##########################################################################################################################################################################
## flowMeans results evaluation

load("ILCFlow.RData")
library(flowMeans)
set.seed(12345)

##########################################################################################################################################################################
## flowMeans loop for nClus
for(nClus in 4:4) {
	
	# parNames2 <- parNames[c(3,4,6,9)]
	flowMeansRes <- flowMeans(inputFCS@exprs, varNames=parNames, #MaxN = NA, 
			NumC = nClus, #iter.max = 50, nstart = 10,
			Mahalanobis = TRUE, 
			Standardize = TRUE, #Update = "Mahalanobis", OrthagonalResiduals=TRUE, MaxCovN=NA, MaxKernN=NA, 
			addNoise=FALSE)     ## flowMeans seems crashing with addNoise=TRUE
	
	resultListFM <- list()
	print("flowMeans results")
	f1VectFM <- c()
	for(refI in 1:length(unique(refLabels))) {
		maxPopF1 <- -1
		MaxPopF1Index <- -1
		# Testing F1 for pop refI, which labels from clustering could that be?
		refLabelsTmp <- ifelse(refLabels == refI, 1, 2)
		for(i in 1:length(unique(flowMeansRes@Label))) {
			labelsTmp <- ifelse(flowMeansRes@Label == i, 1, 2)
			f1 <- f1_score(labelsTmp, refLabelsTmp)[1]
			if (f1 > maxPopF1) {
				maxPopF1 = f1
				MaxPopF1Index = i;
			}
		}
		print(paste("With nClus =", nClus, "best F1 for ref pop", refI, "is based on cluster", MaxPopF1Index, "and = ", maxPopF1))
		resultListFM[[refI]] = list(cluster=MaxPopF1Index, f1=maxPopF1)
		f1VectFM <- cbind(f1VectFM, maxPopF1)
	}
	print(paste("With nClus =", nClus, "flowMeans results mean F1 =", mean(f1VectFM)))
	
	##########################################################################################################################################################################
} ## end of nClus look for flowMeans

save(list = ls(all.names = TRUE), file="ILCFlow.RData")
savehistory(file = "ILCFlow.RHistory")


